@@ -1,21 +1,26 @@
 # (c) 2005 Ian Bicking and contributors; written for Paste (http://pythonpaste.org)
 # Licensed under the MIT license: http://www.opensource.org/licenses/mit-license.php
+from __future__ import with_statement
 import os
-import re
 import sys
-import urllib
-from ConfigParser import ConfigParser
+import re
+
 import pkg_resources
-from paste.deploy.util.fixtypeerror import fix_call
+
+from paste.deploy.compat import ConfigParser, unquote, iteritems
+from paste.deploy.util import fix_call, lookup_object
 
 __all__ = ['loadapp', 'loadserver', 'loadfilter', 'appconfig']
 
+
 ############################################################
 ## Utility functions
 ############################################################
 
+
 def import_string(s):
-    return pkg_resources.EntryPoint.parse("x="+s).load(False)
+    return pkg_resources.EntryPoint.parse("x=" + s).load(False)
+
 
 def _aslist(obj):
     """
@@ -29,6 +34,7 @@ def _aslist(obj):
     else:
         return [obj]
 
+
 def _flatten(lst):
     """
     Flatten a nested list.
@@ -40,11 +46,16 @@ def _flatten(lst):
         result.extend(_flatten(item))
     return result
 
+
 class NicerConfigParser(ConfigParser):
 
     def __init__(self, filename, *args, **kw):
         ConfigParser.__init__(self, *args, **kw)
         self.filename = filename
+        if hasattr(self, '_interpolation'):
+            self._interpolation = self.InterpolateWrapper(self._interpolation)
+
+    read_file = getattr(ConfigParser, 'read_file', ConfigParser.readfp)
 
     def defaults(self):
         """Return the defaults, with their values interpolated (with the
@@ -53,25 +64,49 @@ class NicerConfigParser(ConfigParser):
         Mainly to support defaults using values such as %(here)s
         """
         defaults = ConfigParser.defaults(self).copy()
-        for key, val in defaults.iteritems():
-            defaults[key] = self._interpolate('DEFAULT', key, val, defaults)
+        for key, val in iteritems(defaults):
+            defaults[key] = self.get('DEFAULT', key) or val
         return defaults
 
     def _interpolate(self, section, option, rawval, vars):
+        # Python < 3.2
         try:
             return ConfigParser._interpolate(
                 self, section, option, rawval, vars)
-        except Exception, e:
+        except Exception:
+            e = sys.exc_info()[1]
             args = list(e.args)
-            args[0] = 'Error in file %s, [%s] %s=%r: %s' % (
-                self.filename, section, option, rawval, e)
+            args[0] = 'Error in file %s: %s' % (self.filename, e)
             e.args = tuple(args)
+            e.message = args[0]
             raise
 
+    class InterpolateWrapper(object):
+        # Python >= 3.2
+        def __init__(self, original):
+            self._original = original
+
+        def __getattr__(self, name):
+            return getattr(self._original, name)
+
+        def before_get(self, parser, section, option, value, defaults):
+            try:
+                return self._original.before_get(parser, section, option,
+                                                 value, defaults)
+            except Exception:
+                e = sys.exc_info()[1]
+                args = list(e.args)
+                args[0] = 'Error in file %s: %s' % (parser.filename, e)
+                e.args = tuple(args)
+                e.message = args[0]
+                raise
+
+
 ############################################################
 ## Object types
 ############################################################
 
+
 class _ObjectType(object):
 
     name = None
@@ -80,8 +115,8 @@ class _ObjectType(object):
 
     def __init__(self):
         # Normalize these variables:
-        self.egg_protocols = map(_aslist, _aslist(self.egg_protocols))
-        self.config_prefixes = map(_aslist, _aslist(self.config_prefixes))
+        self.egg_protocols = [_aslist(p) for p in _aslist(self.egg_protocols)]
+        self.config_prefixes = [_aslist(p) for p in _aslist(self.config_prefixes)]
 
     def __repr__(self):
         return '<%s protocols=%r prefixes=%r>' % (
@@ -92,6 +127,7 @@ class _ObjectType(object):
         return fix_call(context.object,
                         context.global_conf, **context.local_conf)
 
+
 class _App(_ObjectType):
 
     name = 'application'
@@ -113,6 +149,7 @@ class _App(_ObjectType):
 
 APP = _App()
 
+
 class _Filter(_ObjectType):
     name = 'filter'
     egg_protocols = [['paste.filter_factory', 'paste.filter_app_factory']]
@@ -134,6 +171,7 @@ class _Filter(_ObjectType):
 
 FILTER = _Filter()
 
+
 class _Server(_ObjectType):
     name = 'server'
     egg_protocols = [['paste.server_factory', 'paste.server_runner']]
@@ -155,6 +193,7 @@ class _Server(_ObjectType):
 
 SERVER = _Server()
 
+
 # Virtual type: (@@: There's clearly something crufty here;
 # this probably could be more elegant)
 class _PipeLine(_ObjectType):
@@ -170,6 +209,7 @@ class _PipeLine(_ObjectType):
 
 PIPELINE = _PipeLine()
 
+
 class _FilterApp(_ObjectType):
     name = 'filter_app'
 
@@ -180,6 +220,7 @@ class _FilterApp(_ObjectType):
 
 FILTER_APP = _FilterApp()
 
+
 class _FilterWith(_App):
     name = 'filtered_with'
 
@@ -196,19 +237,24 @@ class _FilterWith(_App):
 
 FILTER_WITH = _FilterWith()
 
+
 ############################################################
 ## Loaders
 ############################################################
 
+
 def loadapp(uri, name=None, **kw):
     return loadobj(APP, uri, name=name, **kw)
 
+
 def loadfilter(uri, name=None, **kw):
     return loadobj(FILTER, uri, name=name, **kw)
 
+
 def loadserver(uri, name=None, **kw):
     return loadobj(SERVER, uri, name=name, **kw)
 
+
 def appconfig(uri, name=None, relative_to=None, global_conf=None):
     context = loadcontext(APP, uri, name=name,
                           relative_to=relative_to,
@@ -217,6 +263,7 @@ def appconfig(uri, name=None, relative_to=None, global_conf=None):
 
 _loaders = {}
 
+
 def loadobj(object_type, uri, name=None, relative_to=None,
             global_conf=None):
     context = loadcontext(
@@ -224,6 +271,7 @@ def loadobj(object_type, uri, name=None, relative_to=None,
         global_conf=global_conf)
     return context.create()
 
+
 def loadcontext(object_type, uri, name=None, relative_to=None,
                 global_conf=None):
     if '#' in uri:
@@ -247,22 +295,16 @@ def loadcontext(object_type, uri, name=None, relative_to=None,
         uri, path, name=name, relative_to=relative_to,
         global_conf=global_conf)
 
+
 def _loadconfig(object_type, uri, path, name, relative_to,
                 global_conf):
+    isabs = os.path.isabs(path)
     # De-Windowsify the paths:
     path = path.replace('\\', '/')
-    absolute_path = True
-    if sys.platform == 'win32':
-        _absolute_re = re.compile(r'^[a-zA-Z]:')
-        if not _absolute_re.search(path):
-            absolute_path = False
-    else:
-        if not path.startswith('/'):
-            absolute_path = False
-    if not absolute_path:
+    if not isabs:
         if not relative_to:
             raise ValueError(
-                "Cannot resolve relative uri %r; no context keyword "
+                "Cannot resolve relative uri %r; no relative_to keyword "
                 "argument given" % uri)
         relative_to = relative_to.replace('\\', '/')
         if relative_to.endswith('/'):
@@ -271,7 +313,7 @@ def _loadconfig(object_type, uri, path, name, relative_to,
             path = relative_to + '/' + path
     if path.startswith('///'):
         path = path[2:]
-    path = urllib.unquote(path)
+    path = unquote(path)
     loader = ConfigLoader(path)
     if global_conf:
         loader.update_defaults(global_conf, overwrite=False)
@@ -279,6 +321,7 @@ def _loadconfig(object_type, uri, path, name, relative_to,
 
 _loaders['config'] = _loadconfig
 
+
 def _loadegg(object_type, uri, spec, name, relative_to,
              global_conf):
     loader = EggLoader(spec)
@@ -286,10 +329,20 @@ def _loadegg(object_type, uri, spec, name, relative_to,
 
 _loaders['egg'] = _loadegg
 
+
+def _loadfunc(object_type, uri, spec, name, relative_to,
+             global_conf):
+
+    loader = FuncLoader(spec)
+    return loader.get_context(object_type, name, global_conf)
+
+_loaders['call'] = _loadfunc
+
 ############################################################
 ## Loaders
 ############################################################
 
+
 class _Loader(object):
 
     def get_app(self, name=None, global_conf=None):
@@ -317,39 +370,31 @@ class _Loader(object):
             SERVER, name=name, global_conf=global_conf)
 
     _absolute_re = re.compile(r'^[a-zA-Z]+:')
+
     def absolute_name(self, name):
         """
         Returns true if the name includes a scheme
         """
         if name is None:
             return False
-        return self._absolute_re.search(name)        
+        return self._absolute_re.search(name)
+
 
 class ConfigLoader(_Loader):
 
     def __init__(self, filename):
         self.filename = filename = filename.strip()
-        self.parser = NicerConfigParser(self.filename)
-        # Don't lower-case keys:
-        self.parser.optionxform = str
-        # Stupid ConfigParser ignores files that aren't found, so
-        # we have to add an extra check:
-        if not os.path.exists(filename):
-            if filename.strip() != filename:
-                raise OSError(
-                    "File %r not found; trailing whitespace: "
-                    "did you try to use a # on the same line as a filename? "
-                    "(comments must be on their own line)" % filename)
-            raise OSError(
-                "File %r not found" % filename)
-        self.parser.read(filename)
-        self.parser._defaults.setdefault(
-            'here', os.path.dirname(os.path.abspath(filename)))
-        self.parser._defaults.setdefault(
-            '__file__', os.path.abspath(filename))
+        defaults = {
+            'here': os.path.dirname(os.path.abspath(filename)),
+            '__file__': os.path.abspath(filename)
+            }
+        self.parser = NicerConfigParser(filename, defaults=defaults)
+        self.parser.optionxform = str  # Don't lower-case keys
+        with open(filename) as f:
+            self.parser.read_file(f)
 
     def update_defaults(self, new_defaults, overwrite=True):
-        for key, value in new_defaults.items():
+        for key, value in iteritems(new_defaults):
             if not overwrite and key in self.parser._defaults:
                 continue
             self.parser._defaults[key] = value
@@ -436,6 +481,20 @@ class ConfigLoader(_Loader):
             context.global_conf['__file__'] = global_conf['__file__']
         # @@: Should loader be overwritten?
         context.loader = self
+
+        if context.protocol is None:
+            # Determine protocol from section type
+            section_protocol = section.split(':', 1)[0]
+            if section_protocol in ('application', 'app'):
+                context.protocol = 'paste.app_factory'
+            elif section_protocol in ('composit', 'composite'):
+                context.protocol = 'paste.composit_factory'
+            else:
+                # This will work with 'server' and 'filter', otherwise it
+                # could fail but there is an error message already for
+                # bad protocols
+                context.protocol = 'paste.%s_factory' % context_protocol
+
         return context
 
     def _context_from_explicit(self, object_type, local_conf, global_conf,
@@ -542,8 +601,8 @@ class ConfigLoader(_Loader):
                 found.append(name_prefix)
             name = 'main'
         for section in sections:
-            if section.startswith(name_prefix+':'):
-                if section[len(name_prefix)+1:].strip() == name:
+            if section.startswith(name_prefix + ':'):
+                if section[len(name_prefix) + 1:].strip() == name:
                     found.append(section)
         return found
 
@@ -604,6 +663,32 @@ class EggLoader(_Loader):
                 % (name, self.spec, ', '.join(_flatten(protocol_options))))
         return possible[0]
 
+
+class FuncLoader(_Loader):
+    """ Loader that supports specifying functions inside modules, without
+    using eggs at all. Configuration should be in the format:
+        use = call:my.module.path:function_name
+        
+    Dot notation is supported in both the module and function name, e.g.:
+        use = call:my.module.path:object.method
+    """
+    def __init__(self, spec):
+        self.spec = spec
+        if not ':' in spec:
+            raise LookupError("Configuration not in format module:function")
+
+    def get_context(self, object_type, name=None, global_conf=None):
+        obj = lookup_object(self.spec)
+        return LoaderContext(
+            obj,
+            object_type,
+            None, # determine protocol from section type
+            global_conf or {},
+            {},
+            self,
+            )
+
+
 class LoaderContext(object):
 
     def __init__(self, obj, object_type, protocol,
@@ -632,6 +717,7 @@ class LoaderContext(object):
         conf.context = self
         return conf
 
+
 class AttrDict(dict):
     """
     A dictionary that can be assigned to.
